<?php

namespace Verifone\Hosted\Controller\Standard;

use Magento\Customer\Model\ResourceModel\CustomerRepository;
use Magento\Sales\Model\Order;

class Redirect extends \Verifone\Hosted\Controller\Checkout {
	/**
	 * @return \Magento\Framework\App\ResponseInterface|\Magento\Framework\Controller\Result\Json|\Magento\Framework\Controller\ResultInterface
	 * @throws \Magento\Framework\Exception\LocalizedException
	 */
	public function execute() {
		if ( ! $this->getRequest()->isAjax() ) {
			$this->_cancelPayment();
			$this->_checkoutSession->restoreQuote();
			$this->getResponse()->setRedirect(
				$this->getCheckoutHelper()->getUrl( 'checkout' )
			);
		}

		$quote         = $this->getQuote();
		$email         = $this->getRequest()->getParam( 'email' );
		$paymentMethod = $this->getPaymentMethod();
		$quote->setCustomerEmail( $email );

        if(!$quote->getReservedOrderId())
            $quote->reserveOrderId();

        if(!$this->getCustomerSession()->isLoggedIn()){
            $this->getCheckoutSession()->loadCustomerQuote();
            $quote->setCheckoutMethod( \Magento\Checkout\Model\Type\Onepage::METHOD_GUEST );
        }

        $this->getCheckoutHelper()->setUpVerifoneSettings($paymentMethod);
        if($this->getCheckoutHelper()->getVerifoneSettings()->isOrderBeforeRedirect()) {
            try {
                $this->initCheckout();
                $this->_cartManagement->placeOrder(
                    $quote->getId(),
                    $quote->getPayment()
                );
                $order = $this->getOrder();
                $order->setIncrementId($quote->getReservedOrderId())->save();
            } catch (\Exception $e) {
                $this->_logger->error($e->getMessage());
                $this->_logger->error($e->getTraceAsString());
                if (isset($this->messageManager)) {
                    $this->messageManager->addExceptionMessage($e, __('We can\'t place the order.'));
                }
            }
        }

        if ( $this->getCustomerSession()->isLoggedIn() ) {
            $this->getCheckoutSession()->loadCustomerQuote();
            $quote->updateCustomerData( $this->getQuote()->getCustomer() );
            $quote->setCheckoutMethod( \Magento\Checkout\Model\Type\Onepage::METHOD_CUSTOMER );
        }

		$quote->setPaymentMethod( $this->getPaymentMethod()->getCode() );
		$quote->getPayment()->importData( [ 'method' => $this->getPaymentMethod()->getCode() ] );
		$this->_quoteRepository->save( $quote );
		$result = $this->getCheckoutHelper()->processPayment( $paymentMethod, $quote );
		$params = [];

		if ( isset( $result['url'] ) && isset( $result['method'] ) && $result['method'] == "iframe" ) {
			$params['url']  = $result['url'];
		}

		if ( isset( $result['redirect'] ) && isset( $result['method'] ) && $result['method'] == "hpp" ) {
			$params['redirect'] = $result['redirect'];
		}

		$params['result'] = 'success';
		$params['method'] = $result['method'];
		die( json_encode( $params ) );
	}
}
