<?php

namespace Verifone\Hosted\Cron;

use Magento\Framework\Stdlib\DateTime\Timezone,
    Magento\Sales\Model\Order,
    \Psr\Log\LoggerInterface,
    \Magento\Sales\Model\ResourceModel\Order\CollectionFactory;

class CancelPendingOrders {
    const PAYMENT_METHOD = 'verifone_hosted';

    protected $_logger;
    protected CollectionFactory $_orderCollectionFactory;
    protected Timezone $_stdTimezone;

    public function __construct(CollectionFactory $orderCollectionFactory,
                                Timezone          $stdTimezone,
                                LoggerInterface   $logger) {
        $this->_orderCollectionFactory = $orderCollectionFactory;
        $this->_stdTimezone = $stdTimezone;
        $this->_logger = $logger;
    }

    public function execute() {
        $this->_logger->info("Cancel pending orders - starting cron");

        $durationInSeconds = 3600;
        $currentTime = $this->_stdTimezone->date(time() - $durationInSeconds)
            ->format('Y-m-d H:i:s');

        $orders = $this->_orderCollectionFactory
            ->create()
            ->addAttributeToSelect('*')
            ->addFieldToFilter('created_at', ['lt' => $currentTime])
            ->addFieldToFilter('status', ['eq' => 'pending']);

        $orders->getSelect();

        foreach ($orders as $order) {
            if ($order->getPayment()->getMethod() == self::PAYMENT_METHOD) {
                $order->addStatusHistoryComment(__('Order has been canceled automatically'), Order::STATE_CANCELED)
                    ->setIsVisibleOnFront(true)
                    ->setIsCustomerNotified(false)
                    ->setStatus(Order::STATE_CANCELED)
                    ->setState(Order::STATE_CANCELED);

                $order->save();
                $this->_logger->info("Canceled pending order ".$order->getIncrementId());
            }
        }

        $this->_logger->info("Cancel pending orders completed - stopping cron");

        return $this;
    }
}
